/*********************************************************************/
/* welcome.c                                                         */
/* function neblo_welcome(), second part of join phase (requestor)   */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * neblo_welcome:  this peer is welcome in ring, after a previous join req.
 * Returns 0 if success, < 0 otherwise.
 */
int 
neblo_welcome
( reqst_t *reqst )
{
  
  peer_t               *pred,*f;
  peer_t               *succ = NULL;
  u_int32_t             j_handle_val;
  unsigned int          welcport;
  int                   rc;
  char                 *keymod;
  char                 *keyexp;
  ULL                   a;
  unsigned int          fport;
  unsigned int          i;


  _do_(_malloc_(keymod,KEYBUFLEN+SLACK),
       !keymod,"neblo_welcome(): keymod malloc");
  _do_(_malloc_(keyexp,BUFLEN+SLACK),
       !keyexp,"neblo_welcome(): keyexp malloc");

  cmd_handle_lock(&ring->join_handle);

  /*  welcoming peer should eventually become our predecessor in the ring  */
  _do_(_calloc_(pred,sizeof(peer_t)),!pred,"neblo_welcome(): pred calloc");
  _pthmutexinit_(&pred->lock);
  pred->addr = reqst->peer_addr;
  pred->status = TOCALL;

  sscanf(reqst->buf,"%*s %u %u %50000s %100s",
         &j_handle_val,&welcport,keymod,keyexp);
  pred->port = (u_int16_t)welcport;

  /*
   *  SECURITY:  accept only if "welcome" carries the 32 bit handle
   *  originally associated to the join req.
   */

  if (j_handle_val != cmd_handle_val(&ring->join_handle)) {

#ifdef DEBUG_J
    S(fprintf(stderr,"\tneblo_welcome(): invalid join handle\n"));
#endif

    rc = -1;
    goto exit;
  }

  _do_(_RSA_new_(pred->pubkey),!pred->pubkey,
       "neblo_welcome(): RSA_new of pred->pubkey");
  _convert_pubkey_(pred->pubkey,keymod,keyexp,/* */);
  if (!(pred->pubkey->n && pred->pubkey->e)) {

#ifdef DEBUG
    S(fprintf(stderr, "\tbad pubkey from welcoming peer\n"));
#endif

    rc = -1;
    goto exit;
  }

  /*  Welcoming peer put a challenge on our pubkey.  */
  rc = _get_challenge_(reqst->csock,reqst->buf);

  if (rc < 0) {

#ifdef DEBUG_J
    S(fprintf(stderr, "\tchallenge interrupted\n"));
#endif

    goto exit;
  }

#ifdef DEBUG_J
  S(fprintf(stderr, "\tchallenge done\n"));
#endif

  /*  Get upper bound of addr intvl (lower bound is our id. itself).  */

  _read_(reqst->csock,reqst->buf,KEYBUFLEN,rc,TIMEOUT,goto exit);
  if (strncmp(reqst->buf,"intvl ",strlen("intvl "))) {
    rc = -1;
    goto exit;
  }

#ifdef DEBUG_J
  S(fprintf(stderr,"\t>> %s\n",reqst->buf));
#endif

  sscanf(reqst->buf,"%*s %016llx",&a);
  ring->up_bound = a;

#ifdef DEBUG_J
  S(fprintf(stderr, "\tgot addr interval %016llx -- %016llx\n",
            ring->my_id,ring->up_bound));
#endif

  /*  Welcoming peer sends its current neighbour list.  */

  for (i=MAXNEIGH+1;i;i--) {  /* "+1" for the "end" cmd */
    _read_(reqst->csock,reqst->buf,KEYBUFLEN,rc,TIMEOUT,break);
    if (!strncmp(reqst->buf,"end",strlen("end"))) {

#ifdef DEBUG
      S(fprintf(stderr, "\t>> %s\n",reqst->buf));
#endif
      /* no more neighbours to introduce */
      break;
    }
    _do_(_calloc_(f,sizeof(peer_t)),!f,"neblo_welcome(): f calloc");
    _pthmutexinit_(&f->lock);
    sscanf(reqst->buf, "%lu %u %50000s %100s",
           (unsigned long *)&f->addr.s_addr,&fport,keymod,keyexp);
    f->port = (u_int16_t)fport;

#ifdef DEBUG_J
    S(fprintf(stderr,"\t>> \"neighbour: addr %u.%u.%u.%u:%u\"\n",
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port));
#endif

    _do_(_RSA_new_(f->pubkey),!f->pubkey,
         "neblo_welcome(): RSA_new of f->pubkey");

    _convert_pubkey_(f->pubkey,keymod,keyexp,/**/);

    if (!(f->pubkey->n && f->pubkey->e)) {

#ifdef DEBUG
      S(fprintf(stderr, "\tbad neigh pubkey from welcoming peer\n"));
#endif

      _write_(reqst->csock,"ko",strlen("ko")+1,rc,/* nothing */);
      _free_(f);
      rc = -1;
      break;
    }

    f->status = TOCALL;

    /*  All other fields are zero/NULL by calloc  */

    if (i == MAXNEIGH+1) {        /*  first in list --> successor  */
      succ = f;
      /*  Expect a friendship connection from successor  */
      rc = insert_expd_reqst(f->addr,f->port,2 /* from succ */,f->pubkey,
                             reqst->th_id);
      if (rc < 0) {
        _free_(f);
        break;
      }
    }

    append_peer_list(&ring->neigh_list,&ring->neigh_list_tail,&ring->lock,f);

    _write_(reqst->csock,"ok",strlen("ok")+1,rc,break);

#ifdef DEBUG
    S(fprintf(stderr, "\t<< ok\n"));
#endif

  }

  if (rc < 0)
    goto exit;

  /*
   * tell the welcoming peer we are ready to accept its permanent
   * connection with the local peer, aka "friendship".
   * this is necessary in order to get data it will flush to us soon.
   * but, we will accept connections only from him (we only have one
   * single entry in the queue of expected friendship requests).
   * The incoming connection will be flagged as "from predecessor".
   */
  rc = insert_expd_reqst(pred->addr,pred->port,1 /* from predecessor */,
                         pred->pubkey,reqst->th_id);
  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto exit);

#ifdef DEBUG
  S(fprintf(stderr, "\t<< ok\n"));
#endif

  /*
   *  The incoming connection will be managed by a separate thread.
   *  Predecessor tells us "ok" once it has established the conn with
   *  us and is ready to accept the connection in the opposite way.
   */

  bzero(reqst->buf,4);
  _read_(reqst->csock,reqst->buf,BUFLEN,rc,TIMEOUT,goto exit);
  if (strncmp(reqst->buf,"ok",strlen("ok"))) {
    rc = -1;
    goto exit;
  }

#ifdef DEBUG_J
  S(fprintf(stderr, "\t>> %s\n",reqst->buf));
#endif

  /*  Now establish the connection to predecessor  */

  f = NULL;
  forall_friends(ff,if (ff->from_predecessor) {f=ff; break;});

  if (f == NULL) {  /* should not occur */
    rc = -1;
    goto exit;
  }

  for (i=5; i; i--) {
    rc = call_peer(f,2 /* from successor */,0,reqst->th_id);
    if (rc >= 0)
      /* success. predecessor->status is set to OK */
      break;
    if (rc == -2) {
      /* predecessor has died meanwhile. abort... */

#ifdef DEBUG_J
      S(fprintf(stderr,"\tpredecessor dead. abort...\n"));
#endif

      goto exit;
    }
    /* predecessor not responding. wait a moment, then try again */
    sleep(1);
  }
  if (!i) {
    /* predecessor not responding for too long.  assume it is dead */

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tpredecessor not responding, abort...\n",
                reqst->th_id));
#endif

    rc = -1;
    goto exit;
  }

  /*
   * we can expect that our (new) successor will
   * eventually open a connection to us (the backward connection from any
   * peer to its predecessor).
   *
   * Why right here?  Because this is the only place where we are sure
   * it works in all cases, including the (rare but nasty) case when this
   * peer is welcome in a ring counting just one peer (which is predecessor
   * and successor at the same time...that's the nasty thing!).
   */
  f = ring->neigh_list;
  rc = insert_expd_reqst(f->addr,f->port,2 /* from successor */,f->pubkey,
                         reqst->th_id);

  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto exit);

#ifdef DEBUG
  S(fprintf(stderr, "\t<< ok\n"));
#endif

  /* Two-way connection with predecessor has been established. */

  /* get an OK from predecessor when we are allowed to contact our successor */
  bzero(reqst->buf,4);
  _read_(reqst->csock,reqst->buf,BUFLEN,rc,TIMEOUT,goto exit);
  if (strncmp(reqst->buf,"ok",strlen("ok"))) {
    rc = -1;
    goto exit;
  }

#ifdef DEBUG_J
  S(fprintf(stderr, "\t>> %s\n",reqst->buf));
#endif

  /*
   *  The conn. with successor is fundamental and must be opened now,
   *  to ensure that the ring shape is preserved.  Make more attempts if
   *  necessary.
   */
  for (i=5; i; i--) {
    rc = call_peer(succ,1 /* from predecessor */,0,reqst->th_id);
    if (rc >= 0)
      break;
    /*  Successor not yet ready?.  Retry  */
    sleep(1);
  }
  if (rc < 0) {
    /*  Successor unreachable.  We must give up.  */
    rc = -1;
    goto exit;
  }

  /*
   * Connection from this peer to the successor is established.
   * The backward connection from successor will be eventually made,
   * but for the routing correctness we are enough for now.
   * However, this peer will not be allowed to welcome other peers until
   * the backward connection is established as well.
   */

  /* Give a final OK to predecessor: the ring has been sealed */
  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto exit);

#ifdef DEBUG
  S(fprintf(stderr, "\t<< (final) ok\n"));
#endif

  /*  Entered ring successfully :-)  */

#ifdef VERBOSE
  S(fprintf(stdout,"\tring entered :-)\n"));
  S(fprintf(stdout,"\taddr interval: %016llx -- %016llx\n",
            (ULL)ring->my_id,(ULL)ring->up_bound));
  fflush(stdout);
#endif
#ifdef DEBUG
  forall_neighbours(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      S(fprintf(stderr,"neighbour:  %u.%u.%u.%u:%u (%u)\n",
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port,f->status));
  );
#endif

  /* accept incoming connections from all neighbours */
  ring->enabled = 1;
  rc = 0;

exit:

  _free_(keymod); _free_(keyexp);

  if (rc >= 0)
    cmd_handle_reset(&ring->join_handle);
  else {
    clean_peer_list(&ring->neigh_list,&ring->neigh_list_tail,&ring->lock);
    clean_peer_list(&ring->friend_list,&ring->friend_list_tail,&ring->lock);
  }

  cmd_handle_unlock(&ring->join_handle);

  return rc;
  
}
